package org.apache.ojb.broker.metadata;

/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 *    "Apache ObjectRelationalBridge" must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    "Apache ObjectRelationalBridge", nor may "Apache" appear in their name, without
 *    prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

import java.io.Serializable;
import java.util.*;
import java.net.URL;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.ojb.broker.PBFactoryException;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerFactory;
import org.apache.ojb.broker.PersistenceBrokerException;
import org.apache.ojb.broker.PBKey;
import org.apache.ojb.broker.metadata.auto.AutoCreatorDb;

import org.apache.ojb.broker.query.Criteria;
import org.apache.ojb.broker.query.Query;
import org.apache.ojb.broker.query.QueryFactory;

import org.apache.ojb.broker.singlevm.PersistenceBrokerConfiguration;
import org.apache.ojb.broker.util.logging.LoggerFactory;
import org.apache.ojb.broker.util.logging.LoggingHelper;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.configuration.impl.OjbConfigurator;

/**
 * The repository containing all mapping information of the PersistenceBroker.
 * The data is stored in a xml file "repository.xml" in the directory of the main class.
 * @author <a href="mailto:thma@apache.org">Thomas Mahler<a>
 */
public class DescriptorRepository implements Serializable, XmlCapable
{
    /**
     * Default descriptor repository.
     */
    private static String defaultDescriptor;
    
    /**
     * Map manage all descriptor repository instances.
     */
    private static Map repositories = Collections.synchronizedMap(new HashMap());


    /**
     * path to the xml repository file containing the mapping information
     */
    private String _repositoryfile = "repository.xml";

    /**
     * Key to identify this repository instance.
     */
    private PBKey pbKey;

    /**
     * This table holds all known Mapping descriptions.
     * Key values are the respective Class objects
     */
    private Hashtable descriptorTable = new Hashtable();

    /**
     * We need a lot the extent, to which a class belongs
     * (@see DescriptorRepository#getExtentClass). To speed up the costy
     * evaluation, we use this tiny hash map.
     */
    private HashMap extentTable = new HashMap();

    /**
     * The Descriptor of a default JDBC connection, to be used when no
     * individual descriptor is given for a certain class
     */
    private JdbcConnectionDescriptor defaultConnectionDescriptor;

	/**
	 *  Logger for dynamic repository stuff.
	 */
	private Logger logger;

	/**
	 *  Collection of DynamicClassEntry for dynamically loaded classes.
	 */
	private Set dynamicClasses;
	
	/**
	 *  Automatic table creator.
	 */
	private AutoCreatorDb tableCreator;

	private MetadataConfiguration configuration;
		
    //*****************************************************
    // constructor

    /**
     * Constructor declaration
     */
    DescriptorRepository() throws PersistenceBrokerException
    {
    	logger = LoggerFactory.getDefaultLogger();
		configuration = (MetadataConfiguration)OjbConfigurator.getInstance().getConfigurationFor(null);    	
    }

    //*****************************************************
    // static methods

    public static DescriptorRepository getInstance(PBKey pbKey)
        throws MetadataException
    {
        DescriptorRepository ret = (DescriptorRepository) repositories.get(pbKey);
        if(ret == null)
        {
            RepositoryPersistor persistor = new RepositoryPersistor();
            try
            {
                ret = persistor.readFromFile(pbKey.getRepositoryFile());
                prepareJdbcConnectionDescriptor(pbKey, ret.getDefaultJdbcConnection());
                ret.setPBKey(pbKey);
                repositories.put(pbKey, ret);
            }
            catch (Throwable t)
            {
                LoggerFactory.getDefaultLogger().fatal(t);
                throw new MetadataException(t);
            }
        }
        return ret;
    }

    private static void prepareJdbcConnectionDescriptor(PBKey key, JdbcConnectionDescriptor jcd)
    {
        if(key.getUser() != null)
        {
            jcd.setUserName(key.getUser());
            jcd.setPassWord(key.getPassword());
        }
    }

    /**
     *
     */
    public static DescriptorRepository getDefaultInstance() throws MetadataException
    {
        if(defaultDescriptor == null)
        {
            defaultDescriptor = ((PersistenceBrokerConfiguration) OjbConfigurator.getInstance()
                                    .getConfigurationFor(null)).getRepositoryFilename();
        }
        return getInstance(defaultDescriptor);
    }


    /**
     * @deprecated
     * @param repositoryFileName
     * @return the DescriptorRepository instance
     * @throws MetadataException
     */
    public static DescriptorRepository getInstance(String repositoryFileName)
        throws MetadataException
    {
        return getInstance(new PBKey(repositoryFileName, null, null));
    }

    /**
     * returns the singleton instance
     * @deprecated
     * @return DescriptorRepository the singleton instance
     * @exception MetadataException
     */
    public static DescriptorRepository getInstance() throws MetadataException
    {
        if(defaultDescriptor == null)
        {
            defaultDescriptor = ((PersistenceBrokerConfiguration) OjbConfigurator.getInstance()
                                    .getConfigurationFor(null)).getRepositoryFilename();
        }
        return getInstance(defaultDescriptor);
    }


    //*****************************************************
    // methods


    private void setPBKey(PBKey key)
    {
        this.pbKey = key;
    }

    public PBKey getPBkey()
    {
        return this.pbKey;
    }

    /**
     * set the path to the xml repository
     */
    public void setRepositoryPath(String pathToRepository)
    {
        _repositoryfile = pathToRepository;
    }


    /**
     * returns a string representation
     */
    public String toString()
    {
        //return "DescriptorRepository\n" + defaultConnectionDescriptor + "\n" + descriptorTable + "\n" + "";
        return DescriptorRepository.class.getName()+" use repository " + this._repositoryfile;
    }

    /**
     * Method declaration
     *
     *
     * @param jcd
     *
     *
     */
    public void setDefaultJdbcConnection(JdbcConnectionDescriptor jcd)
    {
        defaultConnectionDescriptor = jcd;
    }

    /**
     * Return the {@link JdbcConnectionDescriptor} of the repository.
     */
    public JdbcConnectionDescriptor getDefaultJdbcConnection()
    {
        return defaultConnectionDescriptor;
    }

    /**
     * Add a ClassDescriptor to the internal Hashtable<br>
     * Set the Repository for ClassDescriptor
     */
    public void put(Class c, ClassDescriptor cld)
    {
        cld.setRepository(this); // BRJ
        descriptorTable.put(c, cld);
        Vector extentClasses = cld.getExtentClasses();
        for (int i = 0; i < extentClasses.size(); ++i)
        {
            addExtent((Class) extentClasses.get(i), cld);
        }
    }

    /**
     * Add a pair of extent/classdescriptor to the extentTable to gain speed
     * while retrieval of extents.
     * @param clazz the extent itself
     * @param cld the class descriptor, where it belongs to
     */
    void addExtent(Class clazz, ClassDescriptor cld)
    {
        synchronized (extentTable)
        {
            extentTable.put(clazz, cld);
        }
    }

    /**
     * lookup a ClassDescriptor in the internal Hashtable
     */
    public ClassDescriptor getDescriptorFor(Class c) throws ClassNotPersistenceCapableException
    {
        ClassDescriptor result = resolveClass(c);
        if (result == null)
        {
            throw new ClassNotPersistenceCapableException(c.getName() + " not found in OJB Repository");
        }
        else
        {
            return result;
        }
    }

    public boolean hasDescriptorFor(Class c)
    {
        ClassDescriptor result = resolveClass(c);
        return (result != null);
    }
    
    /**
     * returns the Extent to which the class clazz belongs.
     * This may be a baseclass,an interface or clazz itself, if no Extent
     * is defined.
     * throws a PersistenceBrokerException if clazz is not persistence capable,
     * i.e. if clazz is not defined in the DescriptorRepository.
     * @param the class to lookup the Extent for
     */
    public Class getExtentClass(Class clazz)
    {
        //try to find an extent that contains clazz
        Class extentClass = null;
        synchronized (extentTable)
        {
            ClassDescriptor cld = (ClassDescriptor) extentTable.get(clazz);
            if (cld != null)
            {
                extentClass = cld.getClassOfObject();
            }
        }
        // if such an extent could not be found just return clazz itself.
        if (extentClass == null)
        {
            extentClass = clazz;
        }
        return extentClass;
    }

    public Iterator iterator()
    {
        return descriptorTable.values().iterator();
    }

    /*
     * @see XmlCapable#toXML()
     */
    public String toXML()
    {
        RepositoryTags tags = RepositoryTags.getInstance();
        String eol = System.getProperty("line.separator");

        // 1. write XML header
        String strReturn = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + eol;
        strReturn += "<!-- OJB generated this file on " + new Date().toString() + " -->" + eol;
        strReturn += "<!DOCTYPE MappingRepository SYSTEM \"repository.dtd\">" + eol;
        strReturn += tags.getOpeningTagById(MappingRepository) + eol;

        // 2. write default JdbcConnectionDescriptor
        strReturn += this.getDefaultJdbcConnection().toXML();

        // 3. write all ClassDescriptors
        Iterator i = this.iterator();
        while (i.hasNext())
        {
            strReturn += ((XmlCapable) i.next()).toXML() + eol;
        }
        // 4. write closing tag
        return strReturn + tags.getClosingTagById(MappingRepository);

    }

    /**
     * @return all field descriptors for a class that belongs to a set of classes mapped
     * to the same table, otherwise the select queries produced won't contain the necessary
     * information to materialize extents mapped to the same class.
     */
    public Collection getFieldDescriptorsForMultiMappedTable(ClassDescriptor targetCld)
    {
        return getAllMappedColumns(getClassesMappedToSameTable(targetCld));
    }

    private Collection getAllMappedColumns(List classDescriptors)
    {
        Iterator it = classDescriptors.iterator();
        HashMap map = new HashMap();
        ClassDescriptor temp = null;
        while (it.hasNext())
        {
            temp = (ClassDescriptor) it.next();
            FieldDescriptor[] fields = temp.getFieldDescriptions();
            if (fields != null)
            {
                for (int i = 0; i < fields.length; i++)
                {
                    /**
                     * MBAIRD
                     * hashmap will only allow one entry per unique key, so no need to check contains(fields[i].getColumnName()).
                     */
                    map.put(fields[i].getColumnName(), fields[i]);
                }
            }
        }
        return map.values();
    }

    private List getClassesMappedToSameTable(ClassDescriptor targetCld)
    {
        //try to find an extent that contains clazz
        Iterator iter = descriptorTable.values().iterator();
        List retval = new ArrayList();
        while (iter.hasNext())
        {
            ClassDescriptor cld = (ClassDescriptor) iter.next();
            if (cld.getFullTableName() != null)
            {
                if (cld.getFullTableName().equals(targetCld.getFullTableName()))
                {
                    retval.add(cld);
                }
            }
        }
        return retval;
    }

    public Hashtable getDescriptorTable(){
        return descriptorTable;
    }

    public void printDescriptorTable(){
        Enumeration enum = descriptorTable.keys();
        Enumeration enum2 = descriptorTable.elements();
        while (enum.hasMoreElements()) {
            Object cl =  enum.nextElement();
            ClassDescriptor descriptor  = (ClassDescriptor)enum2.nextElement();
            System.out.println(cl + " - " + descriptor.getFullTableName());
        }
    }
    
    
    // Additions by Martin Fuzzey from here
    
    /**
     * Obtain a class descriptor for the given class.
     * Initally do a look up in the map (loaded from the repository.xml)
     * If not found attempt to obtain from the classpath (as classname-OJB.xml)
     * If found cache in map for subsequent lookups.
     *
     * @return class descriptor or null if not found
     */
    private ClassDescriptor resolveClass(Class c) {
        ClassDescriptor result = (ClassDescriptor) descriptorTable.get(c);
        if (result != null) return result;
        
        // Obtain the dynamic classes memorized from previous OJB
        // invocations. This is a once only action done here rather
        // than during creation to :
        // a) avoid chicken and egg situation with default PB instance
        // b) avoid overhead if no dynamic classes ever used.
        // Note that thiss means that if an extent is defined in the
        // static repository its dynamic memebers will not be available
        // for queries until at least one dynamic class has been loaded...
        if (loadDynamicClassesIfRequired()) {
        	// if we have loaded we may have obtained the class we're looking for
        	// check and return here if so since reloading the descriptor
        	// from XML for an extent will nuke the extent info
	        result = (ClassDescriptor) descriptorTable.get(c);
    	    if (result != null) {
    	    	return result;
    	    }
        }
        
        String descName = getDescriptorResourceName(c);
        URL url = c.getResource(descName);
        if (url == null) {
            logger.debug("DYN: No dynamic repository resource for " + c);
            return null;
        }
        try {
        	logger.info("DYN: Attempting dynamic repository load for " + c + " from " + url);
            getRepositoryPersistor().appendRepository(this, url);
        } catch(Throwable ex) {
        	logger.error("DYN: error loading dyanmic XML from " + url, ex);
        }
        result = (ClassDescriptor) descriptorTable.get(c);
        if (result == null) {
            logger.info("DYN: Repository Not found for " + c + " after dynamic load");
            return null;
        }

        // Handle extents
        Class c1 = getExtentClass(c);        
        if (c.equals(c1)) { // only look if we have no extents
        	logger.debug("DYN: Looking for extents for " + c);
            Class superClass = c.getSuperclass();
			boolean extentFound = resolveExtent(c, superClass);            
			
          	// Try all interfaces (both direct and from super classes)
        	Class itfClass;
        	for (itfClass = c; 
        			!extentFound && itfClass != null && !itfClass.equals(Object.class); 
        			itfClass = c.getSuperclass()) {            		
                Class[] interfaces = itfClass.getInterfaces();
                for (int i=0; !extentFound && i<interfaces.length; i++) {
                	extentFound = resolveExtent(c, interfaces[i]);
    	        }
        	}
        }
        
        if (result != null) {
        	storeDynamicClassIfRequired(c);
        	createTableIfRequired(result);
        }
        return result;
    }

	/**
	 * Obtain a persistor.
	 * This implementation just creates a new instance but could be replaced...
	 */
	private RepositoryPersistor getRepositoryPersistor() {
		return new RepositoryPersistor();
	}

    
    private boolean resolveExtent(Class base, Class extent) {
        boolean found = false;
        if (extent != null && !extent.equals(Object.class)) {
            ClassDescriptor extentDesc = resolveClass(extent);
            if (extentDesc != null) {
                addExtent(base, extentDesc);
                extentDesc.addExtentClass(base);
                logger.info("DYN: Found extent " + extent + " for " + base);
                found = true;
            }
        }
        return found;
    }

	/**
	 * Obtain the dynamically loaded classes from persistant storage
	 * if not already done.
	 * @return true if a load was done
	 */
	private boolean loadDynamicClassesIfRequired() {
		boolean done = false;
		synchronized(this) {
			if (dynamicClasses == null) {
				dynamicClasses = new HashSet();
				loadDynamicClasses();
				done = true;
			}
		}
		return done;
	}

	/**
	 * Obtain the dynamically loaded classes from persistant storage.
	 */
	private void loadDynamicClasses() {			
		PersistenceBroker broker = null;
		try {
			broker = PersistenceBrokerFactory.defaultPersistenceBroker();
		} catch (PBFactoryException e) {
			logger.error(e);
			return;
		}
    	Criteria crit = null;
        Query q = QueryFactory.newQuery(DynamicClassEntry.class, crit);
        
        // obtain dynamically loaded classnames from DB
        Collection dynEntries = broker.getCollectionByQuery(q);
		if (dynEntries != null ) {
			logger.info("DYN: Loaded " + dynEntries.size() + " classes from persist");
		} else {
			logger.info("DYN: No classes in persist");
			return;
		}
		
		// add all to the memory list to prevent unneeded store attempts
		for (Iterator it = dynEntries.iterator(); it.hasNext();) {
			DynamicClassEntry entry = (DynamicClassEntry)it.next();
			String className = entry.getClassName();
			dynamicClasses.add(className);
		}
				
		// resolve them
		for (Iterator it = dynamicClasses.iterator(); it.hasNext();) {
			String className = (String)it.next();
			try {
				Class c = Class.forName(className);
				logger.debug("DYN: Loaded class " + className + " from persist");
				resolveClass(c);
			} catch (ClassNotFoundException ex) {
				logger.error("DYN: cannot load " + className, ex);
			}
		}
	}    
	
	/**
	 * Add a dynamically loaded class to the persistant list.
	 */
	private void storeDynamicClassIfRequired(Class c) {
		String className = c.getName();
		synchronized(dynamicClasses) {
			if (!dynamicClasses.contains(className)) {
				DynamicClassEntry entry = new DynamicClassEntry(className);
				PersistenceBroker broker;
				try {
					logger.info("DYN: Writing " + c + " to persist");
					broker = PersistenceBrokerFactory.defaultPersistenceBroker();
					broker.store(entry);		
				} catch (PBFactoryException e) {
					logger.error(e);
				}
				dynamicClasses.add(className);
			}
		}
	}

	/**
	 * Create tables for class descriptor if required.
	 */
	private void createTableIfRequired(ClassDescriptor cld) {
		if (configuration.allowAutocreateDynamicTables()) {
			try {
				AutoCreatorDb creator = getTableCreator();
				creator.createIfRequiredForClass(cld);
			} catch (Exception ex) {
				logger.error("Failed to autocreate for " + cld, ex);
				// ignore errors (fatal error may occur later...)
			}
		}
	}
	
	private AutoCreatorDb getTableCreator() throws PBFactoryException  {
		synchronized(this) {
			if (tableCreator == null) {
				PersistenceBroker broker = PersistenceBrokerFactory.defaultPersistenceBroker();				
				tableCreator = new AutoCreatorDb(broker);
			}
		}
		return tableCreator;
	}
		
    /**
     * Obtain the descriptor resource name for a given class.
     * This implementation uses className-OJB.xml
     */
    protected String getDescriptorResourceName(Class c) {
        StringBuffer sb = new StringBuffer();
        String name = c.getName();
        int idx = name.lastIndexOf(".");
        
        if (idx == -1) {
            sb.append(name); // no package
        } else {
            sb.append(name.substring(idx + 1));
        }
        sb.append("-OJB.xml");
        return sb.toString();
    }
    
    /**
     * Reset the mapping, forgets any dynamically loaded classes.
     * Intended for test code to simulate a new OJB startup (doing
     * a close() on the PB does not work since the PB instances
     * are now pooled together with their descriptors.
     */
    public void reset() throws MetadataException {
    	logger.info("DYN: Resetting repository " + pbKey);
    	descriptorTable.clear();
    	extentTable.clear();
    	dynamicClasses.clear();
    	dynamicClasses = null;  // force reload
    	try {
	    	getRepositoryPersistor().appendRepository(this, pbKey.getRepositoryFile());
    	} catch (Exception ex) {
        	LoggerFactory.getDefaultLogger().fatal(ex);
            throw new MetadataException(ex);
    	}
    }
}
