package org.apache.ojb.broker.metadata;

/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 *    "Apache ObjectRelationalBridge" must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    "Apache ObjectRelationalBridge", nor may "Apache" appear in their name, without
 *    prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.ojb.broker.PersistenceBrokerException;
import org.apache.ojb.broker.accesslayer.RowReader;
import org.apache.ojb.broker.accesslayer.conversions.FieldConversion;
import org.apache.ojb.broker.util.logging.Logger;
import org.apache.ojb.broker.util.logging.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
    
/**
 * The handler catches Parsing events raised by the xml-parser
 * and builds up the DescriptorRepository that is used within the
 * OJB PersistenceBroker System.
 *
 * @author Thomas Mahler
 * @author Jakob Braeuchli
 */
public class RepositoryXmlHandler
    extends DefaultHandler
    implements RepositoryElements, IsolationLevels
{
    private Logger logger;

    private DescriptorRepository m_repository;
    private JdbcConnectionDescriptor m_CurrentJCD;
    private ClassDescriptor m_CurrentCLD;
    private FieldDescriptor m_CurrentFLD;
    private ObjectReferenceDescriptor m_CurrentORD;
    private CollectionDescriptor m_CurrentCOD;
    private String m_CurrentString;

    /** the default isolation level*/
    private int defIsoLevel = IL_DEFAULT;

    /**
     * All known xml tags are kept in this table.
     * The tags table allows lookup from literal to id
     * and from id to literal.
     */
    private RepositoryTags tags = RepositoryTags.getInstance();

    /**
     * returns the XmlCapable id associated with the literal.
     * OJB maintains a RepositoryTags table that provides
     * a mapping from xml-tags to XmlCapable ids.
     * @param literal the literal to lookup
     * @return the int value representing the XmlCapable
     *
     * @throws MetadataException if no literal was found in tags mapping
     */
    private int getLiteralId(String literal) throws PersistenceBrokerException
    {
        //logger.debug("lookup: " + literal);
        try
        {
            return tags.getIdByTag(literal);
        }
        catch (NullPointerException t)
        {
            throw new MetadataException("unknown literal: '" + literal + "'",t);
        }

    }

    /**
     * build a handler that fills the given repository
     * from an XML file.
     */
    public RepositoryXmlHandler(DescriptorRepository dr)
    {
    	if (dr != null)
    	{
        	m_repository = dr;
    	}
        else
        {
        	m_repository = DescriptorRepository.getDefaultInstance();
        }
        logger = LoggerFactory.getLogger(this.getClass());
    }

    /**
     * startDocument callback, nothing to do here.
     */
    public void startDocument()
    {
        logger.debug("startDoc");
    }

    /**
     * endDocument callback, nothing to do here.
     */
    public void endDocument()
    {
        logger.debug("endDoc");
    }

    /**
     * startElement callback.
     * Only some Elements need special start operations.
     * @throws MetadataException indicating mapping errors
     */
    public void startElement(String uri, String name, String qName, Attributes atts)
    {
        m_CurrentString = null;
        try
        {
            switch (getLiteralId(qName))
            {
                case MappingRepository :
                    {
                        String defIso = atts.getValue(tags.getTagById(ISOLATION_LEVEL));
                        if (defIso != null)
                        {
                            defIsoLevel = getIsoLevel(defIso);
                        }
                        logger.debug(" > " + tags.getTagById(MappingRepository));

                        break;
                    }
                case JdbcConnectionDescriptor :
                    {
                        logger.debug("   > " + tags.getTagById(JdbcConnectionDescriptor));
                        m_CurrentJCD = new JdbcConnectionDescriptor();
                        if (m_CurrentCLD == null)
                        {
                            m_repository.setDefaultJdbcConnection(m_CurrentJCD);
                        }
                        else
                        {
                            m_CurrentCLD.setConnectionDescriptor(m_CurrentJCD);
                        }
                        // set platform attribute
                        String platform = atts.getValue(tags.getTagById(Dbms_name));
                        logger.debug("     " + tags.getTagById(Dbms_name) + ": " + platform);
                        m_CurrentJCD.setDbms(platform);

                        // set jdbc-level attribute
                        String level = atts.getValue(tags.getTagById(Jdbc_level));
                        logger.debug("     " + tags.getTagById(Jdbc_level) + ": " + level);
                        m_CurrentJCD.setJdbcLevel(level);

                        // set driver attribute
                        String driver = atts.getValue(tags.getTagById(Driver_name));
                        logger.debug("     " + tags.getTagById(Driver_name) + ": " + driver);
                        m_CurrentJCD.setDriver(driver);

                        // set protocol attribute
                        String protocol = atts.getValue(tags.getTagById(Url_protocol));
                        logger.debug("     " + tags.getTagById(Url_protocol) + ": " + protocol);
                        m_CurrentJCD.setProtocol(protocol);

                        // set subprotocol attribute
                        String subprotocol = atts.getValue(tags.getTagById(Url_subprotocol));
                        logger.debug("     " + tags.getTagById(Url_subprotocol) + ": " + subprotocol);
                        m_CurrentJCD.setSubProtocol(subprotocol);

                        // set the dbalias attribute
                        String dbalias = atts.getValue(tags.getTagById(Url_dbalias));
                        logger.debug("     " + tags.getTagById(Url_dbalias) + ": " + dbalias);
                        m_CurrentJCD.setDbAlias(dbalias);

                        // set the datasource attribute
                        String datasource = atts.getValue(tags.getTagById(Datasource_name));
                        logger.debug("     " + tags.getTagById(Datasource_name) + ": " + datasource);
                        m_CurrentJCD.setDatasourceName(datasource);

                        // set the user attribute
                        String user = atts.getValue(tags.getTagById(User_name));
                        logger.debug("     " + tags.getTagById(User_name) + ": " + user);
                        m_CurrentJCD.setUserName(user);

                        // set the password attribute
                        String password = atts.getValue(tags.getTagById(User_passwd));
                        logger.debug("     " + tags.getTagById(User_passwd) + ": " + password);
                        m_CurrentJCD.setPassWord(password);

                        break;
                    }
                case ClassDescriptor :
                    {
                        logger.debug("  > " + tags.getTagById(ClassDescriptor));
                        m_CurrentCLD = new ClassDescriptor(m_repository);

                        // set isolation-level attribute
                        String isoLevel = atts.getValue(tags.getTagById(ISOLATION_LEVEL));
                        logger.debug("     " + tags.getTagById(ISOLATION_LEVEL) + ": " + isoLevel);
                        m_CurrentCLD.setIsolationLevel(getIsoLevel(isoLevel));

                        // set class attribute
                        String classname = atts.getValue(tags.getTagById(Class_name));
                        logger.debug("     " + tags.getTagById(Class_name) + ": " + classname);
                        m_CurrentCLD.setIsolationLevel(getIsoLevel(isoLevel));
                        try
                        {
                            Class c = Class.forName(classname, true, Thread.currentThread().getContextClassLoader());
                            m_CurrentCLD.setClassOfObject(c);
                            m_repository.put(c, m_CurrentCLD);
                        }
                        catch (Exception ex)
                        {
                            logger.error(ex);
                            throw new PersistenceBrokerException(ex);
                        }

                        // set schema attribute
                        String schema = atts.getValue(tags.getTagById(Schema_name));
                        if (schema != null)
                        {
                            logger.debug("     " + tags.getTagById(Schema_name) + ": " + schema);
                            m_CurrentCLD.setSchema(schema);
                        }

                        // set proxy attribute
                        String proxy = atts.getValue(tags.getTagById(Class_proxy));
                        logger.debug("     " + tags.getTagById(Class_proxy) + ": " + proxy);
                        if (proxy != null)
                        {
                            if (proxy.equals("dynamic"))
                            {
                                Class pc = getDynamicProxyClass(m_CurrentCLD);
                                m_CurrentCLD.setProxyClass(pc);
                            }
                            else
                            {
                                Class pc = Class.forName(proxy, true, Thread.currentThread().getContextClassLoader());
                                m_CurrentCLD.setProxyClass(pc);
                            }
                        }

                        // set table attribute:
                        String table = atts.getValue(tags.getTagById(Table_name));
                        logger.debug("     " + tags.getTagById(Table_name) + ": " + table);
                        m_CurrentCLD.setTableName(table);
                        if (table == null)
                        {
                            m_CurrentCLD.setIsInterface(true);
                        }

                        // set row-reader attribute
                        String rowreader = atts.getValue(tags.getTagById(RowReader));
                        logger.debug("     " + tags.getTagById(RowReader) + ": " + rowreader);
                        if (rowreader != null)
                        {
                            m_CurrentCLD.setRowReader(
                                (RowReader) Class
                                    .forName(rowreader, true, Thread.currentThread().getContextClassLoader())
                                    .newInstance());
                        }

                        break;
                    }

                case Class_extent :
                    {
                        String classname = atts.getValue("class-ref");
                        logger.debug("     " + tags.getTagById(Class_extent) + ": " + classname);
                        Class ec = Class.forName(classname, true, Thread.currentThread().getContextClassLoader());
                        m_CurrentCLD.addExtentClass(ec);
                        break;
                    }

                case FIELDDESCRIPTOR :
                    {
                        logger.debug("    > " + tags.getTagById(FIELDDESCRIPTOR));

                        int id = Integer.parseInt(atts.getValue("id"));
                        m_CurrentFLD = new FieldDescriptor(id);
                        m_CurrentCLD.addFieldDescriptor(m_CurrentFLD);

                        String fieldName = atts.getValue(tags.getTagById(Field_name));
                        logger.debug("     " + tags.getTagById(Field_name) + ": " + fieldName);
                        Class c = m_CurrentCLD.getClassOfObject();
                        m_CurrentFLD.setPersistentField(c, fieldName);

                        String columnName = atts.getValue(tags.getTagById(Column_name));
                        logger.debug("     " + tags.getTagById(Column_name) + ": " + columnName);
                        m_CurrentFLD.setColumnName(columnName);

                        String jdbcType = atts.getValue(tags.getTagById(Jdbc_type));
                        logger.debug("     " + tags.getTagById(Jdbc_type) + ": " + jdbcType);
                        m_CurrentFLD.setColumnType(jdbcType);

                        String primaryKey = atts.getValue(tags.getTagById(PrimaryKey));
                        logger.debug("     " + tags.getTagById(PrimaryKey) + ": " + primaryKey);
                        boolean b = (new Boolean(primaryKey)).booleanValue();
                        m_CurrentFLD.setPrimaryKey(b);

                        String nullable = atts.getValue(tags.getTagById(NULLABLE));
                        logger.debug("     " + tags.getTagById(NULLABLE) + ": " + nullable);
                        b = (new Boolean(nullable)).booleanValue();
                        m_CurrentFLD.setRequired(b);

                        String indexed = atts.getValue(tags.getTagById(INDEXED));
                        logger.debug("     " + tags.getTagById(INDEXED) + ": " + indexed);
                        b = (new Boolean(indexed)).booleanValue();
                        m_CurrentFLD.setAutoIncrement(b);

                        String autoincrement = atts.getValue(tags.getTagById(Autoincrement));
                        logger.debug("     " + tags.getTagById(Autoincrement) + ": " + autoincrement);
                        b = (new Boolean(autoincrement)).booleanValue();
                        m_CurrentFLD.setAutoIncrement(b);

                        String locking = atts.getValue(tags.getTagById(Locking));
                        logger.debug("     " + tags.getTagById(Locking) + ": " + locking);
                        b = (new Boolean(locking)).booleanValue();
                        m_CurrentFLD.setLocking(b);

                        String fieldConversion = atts.getValue(tags.getTagById(FieldConversion));
                        logger.debug("     " + tags.getTagById(FieldConversion) + ": " + fieldConversion);
                        if (fieldConversion != null)
                        {
                            m_CurrentFLD.setFieldConversion(
                                (FieldConversion) Class
                                    .forName(fieldConversion, true, Thread.currentThread().getContextClassLoader())
                                    .newInstance());
                        }

                        // set length attribute
                        String length = atts.getValue(tags.getTagById(LENGTH));
                        if (length != null)
                        {
                            int i = Integer.parseInt(length);
                            logger.debug("     " + tags.getTagById(LENGTH) + ": " + i);
                            m_CurrentFLD.setLength(i);
                            m_CurrentFLD.setLengthSpecified(true);
                        }

                        // set precision attribute
                        String precision = atts.getValue(tags.getTagById(PRECISION));
                        if (precision != null)
                        {
                            int i = Integer.parseInt(precision);
                            logger.debug("     " + tags.getTagById(PRECISION) + ": " + i);
                            m_CurrentFLD.setPrecision(i);
                            m_CurrentFLD.setPrecisionSpecified(true);
                        }

                        // set scale attribute
                        String scale = atts.getValue(tags.getTagById(SCALE));
                        if (scale != null)
                        {
                            int i = Integer.parseInt(scale);
                            logger.debug("     " + tags.getTagById(SCALE) + ": " + i);
                            m_CurrentFLD.setScale(i);
                            m_CurrentFLD.setScaleSpecified(true);
                        }

                        break;
                    }
                case ReferenceDescriptor :
                    {
                        logger.debug("    > " + tags.getTagById(ReferenceDescriptor));
                        m_CurrentORD = new ObjectReferenceDescriptor();
                        m_CurrentCLD.addObjectReferenceDescriptor(m_CurrentORD);

                        // set name attribute
                        name = atts.getValue(tags.getTagById(Field_name));
                        logger.debug("     " + tags.getTagById(Field_name) + ": " + name);
                        m_CurrentORD.setPersistentField(m_CurrentCLD.getClassOfObject(), name);

                        // set class-ref attribute
                        String classRef = atts.getValue(tags.getTagById(Referenced_class));
                        logger.debug("     " + tags.getTagById(Referenced_class) + ": " + classRef);
                        m_CurrentORD.setItemClass(
                            Class.forName(classRef, true, Thread.currentThread().getContextClassLoader()));

                        // set proxy attribute
                        String proxy = atts.getValue(tags.getTagById(ProxyReference));
                        logger.debug("     " + tags.getTagById(ProxyReference) + ": " + proxy);
                        boolean b = (new Boolean(proxy)).booleanValue();
                        m_CurrentORD.setLazy(b);

                        // set refresh attribute
                        String refresh = atts.getValue(tags.getTagById(RefreshReference));
                        logger.debug("     " + tags.getTagById(RefreshReference) + ": " + refresh);
                        b = (new Boolean(refresh)).booleanValue();
                        m_CurrentORD.setRefresh(b);

                        // set auto-retrieve attribute
                        String autoRetrieve = atts.getValue(tags.getTagById(Auto_retrieve));
                        logger.debug("     " + tags.getTagById(Auto_retrieve) + ": " + autoRetrieve);
                        b = (new Boolean(autoRetrieve)).booleanValue();
                        m_CurrentORD.setCascadeRetrieve(b);

                        // set auto-update attribute
                        String autoUpdate = atts.getValue(tags.getTagById(Auto_update));
                        logger.debug("     " + tags.getTagById(Auto_update) + ": " + autoUpdate);
                        b = (new Boolean(autoUpdate)).booleanValue();
                        m_CurrentORD.setCascadeStore(b);

                        //set auto-delete attribute
                        String autoDelete = atts.getValue(tags.getTagById(Auto_delete));
                        logger.debug("     " + tags.getTagById(Auto_delete) + ": " + autoDelete);
                        b = (new Boolean(autoDelete)).booleanValue();
                        m_CurrentORD.setCascadeDelete(b);

                        break;
                    }

                case FOREIGN_KEY :
                    {
                        logger.debug("    > " + tags.getTagById(FOREIGN_KEY));
                        String fieldIdRef = atts.getValue(tags.getTagById(Fk_descriptor_id));
                        logger.debug("      " + tags.getTagById(Fk_descriptor_id) + ": " + fieldIdRef);

                        try
                        {
                            int fieldId;
                            fieldId = Integer.parseInt(fieldIdRef);
                            m_CurrentORD.addForeignKeyField(fieldId);
                        }
                        catch (NumberFormatException rex)
                        {
                            throw new MetadataException(tags.getTagById(Fk_descriptor_id)
                                                        + " attribute must be an int. Found: "
                                                        + fieldIdRef,rex);
                        }
                        break;
                    }

                case CollectionDescriptor :
                    {
                        logger.debug("    > " + tags.getTagById(CollectionDescriptor));
                        m_CurrentCOD = new CollectionDescriptor();
                        m_CurrentCLD.addCollectionDescriptor(m_CurrentCOD);

                        //////
                        // set name attribute
                        name = atts.getValue(tags.getTagById(Field_name));
                        logger.debug("     " + tags.getTagById(Field_name) + ": " + name);
                        m_CurrentCOD.setPersistentField(m_CurrentCLD.getClassOfObject(), name);

                        // set collection-class attribute
                        String collectionClass = atts.getValue(tags.getTagById(Collection_class));
                        logger.debug("     " + tags.getTagById(Collection_class) + ": " + collectionClass);
                        if (collectionClass != null)
                        {
                            m_CurrentCOD.setCollectionClass(
                                Class.forName(collectionClass, true, Thread.currentThread().getContextClassLoader()));
                        }
                        // set element-class-ref attribute
                        String elementClassRef = atts.getValue(tags.getTagById(Items_class));
                        logger.debug("     " + tags.getTagById(Items_class) + ": " + elementClassRef);
                        if (elementClassRef != null)
                        {
                            m_CurrentCOD.setItemClass(
                                Class.forName(elementClassRef, true, Thread.currentThread().getContextClassLoader()));
                        }

                        //set orderby and sort attributes:
                        String orderby = atts.getValue(tags.getTagById(Orderby));
                        String sort = atts.getValue(tags.getTagById(SORT));
                        logger.debug("     " + tags.getTagById(SORT) + ": " + orderby + ", " + sort);

                        if (orderby != null)
                        {
                            boolean isAsc = true;
                            if ("DESC".equalsIgnoreCase(sort))
                            {
                                isAsc = false;
                            }
                            m_CurrentCOD.setOrderBy(orderby);
                            m_CurrentCOD.setAscending(isAsc);
                        }

                        // set indirection-table attribute
                        String indirectionTable = atts.getValue(tags.getTagById(Indirection_table));
                        logger.debug("     " + tags.getTagById(Indirection_table) + ": " + indirectionTable);
                        m_CurrentCOD.setIndirectionTable(indirectionTable);

                        // set proxy attribute
                        String proxy = atts.getValue(tags.getTagById(ProxyReference));
                        logger.debug("     " + tags.getTagById(ProxyReference) + ": " + proxy);
                        boolean b = (new Boolean(proxy)).booleanValue();
                        m_CurrentCOD.setLazy(b);

                        // set refresh attribute
                        String refresh = atts.getValue(tags.getTagById(RefreshReference));
                        logger.debug("     " + tags.getTagById(RefreshReference) + ": " + refresh);
                        b = (new Boolean(refresh)).booleanValue();
                        m_CurrentCOD.setRefresh(b);

                        // set auto-retrieve attribute
                        String autoRetrieve = atts.getValue(tags.getTagById(Auto_retrieve));
                        logger.debug("     " + tags.getTagById(Auto_retrieve) + ": " + autoRetrieve);
                        b = (new Boolean(autoRetrieve)).booleanValue();
                        m_CurrentCOD.setCascadeRetrieve(b);

                        // set auto-update attribute
                        String autoUpdate = atts.getValue(tags.getTagById(Auto_update));
                        logger.debug("     " + tags.getTagById(Auto_update) + ": " + autoUpdate);
                        b = (new Boolean(autoUpdate)).booleanValue();
                        m_CurrentCOD.setCascadeStore(b);

                        //set auto-delete attribute
                        String autoDelete = atts.getValue(tags.getTagById(Auto_delete));
                        logger.debug("     " + tags.getTagById(Auto_delete) + ": " + autoDelete);
                        b = (new Boolean(autoDelete)).booleanValue();
                        m_CurrentCOD.setCascadeDelete(b);

                        break;
                    }
                case Inverse_fk :
                    {
                        logger.debug("    > " + tags.getTagById(Inverse_fk));
                        String fieldIdRef = atts.getValue(tags.getTagById(Fk_descriptor_id));
                        logger.debug("      " + tags.getTagById(Fk_descriptor_id) + ": " + fieldIdRef);

                        try
                        {
                            int fieldId;
                            fieldId = Integer.parseInt(fieldIdRef);
                            m_CurrentCOD.addForeignKeyField(fieldId);
                        }
                        catch (NumberFormatException rex)
                        {
                            throw new MetadataException(tags.getTagById(Fk_descriptor_id)
                                                        + " attribute must be an int. Found: "
                                                        + fieldIdRef,rex);
                        }
                        break;
                    }

                case Fk_pointing_to_this_class :
                    {
                        logger.debug("    > " + tags.getTagById(Fk_pointing_to_this_class));
                        String column = atts.getValue("column");
                        logger.debug("      " + "column" + ": " + column);
                        m_CurrentCOD.addFkToThisClass(column);
                        break;
                    }

                case Fk_pointing_to_items_class :
                    {
                        logger.debug("    > " + tags.getTagById(Fk_pointing_to_items_class));
                        String column = atts.getValue("column");
                        logger.debug("      " + "column" + ": " + column);
                        m_CurrentCOD.addFkToItemClass(column);
                        break;
                    }
                    /*
                    conf.maxActive = cpd.getMaxActive();
                    conf.maxIdle = cpd.getMaxIdle();
                    conf.maxWait = cpd.getMaxWait();
                    conf.minEvictableIdleTimeMillis = cpd.getMinEvictableIdleTimeMillis();
                    conf.numTestsPerEvictionRun = cpd.getNumTestsPerEvictionRun();
                    conf.testOnBorrow = true;//this have always to be true
                    conf.testOnReturn = false;
                    conf.testWhileIdle = false;
                    conf.timeBetweenEvictionRunsMillis = cpd.getTimeBetweenEvictionRunsMillis();
                    conf.whenExhaustedAction = cpd.getWhenExhaustedAction();
                    */
                case connection_pool :
                    {
                        logger.debug("    > " + tags.getTagById(connection_pool));
                        ConnectionPoolDescriptor m_CurrentCPD = m_CurrentJCD.getConnectionPoolDescriptor();
                        m_CurrentJCD.setConnectionPoolDescriptor(m_CurrentCPD);

                        String maxActive = atts.getValue(tags.getTagById(conMaxActive));
                        logger.debug("     " + tags.getTagById(conMaxActive) + ": " + maxActive);
                        if(checkString(maxActive)) m_CurrentCPD.setMaxActive(new Integer(maxActive).intValue());

                        String maxIdle = atts.getValue(tags.getTagById(conMaxIdle));
                        logger.debug("     " + tags.getTagById(conMaxIdle) + ": " + maxIdle);
                        if(checkString(maxIdle)) m_CurrentCPD.setMaxIdle(new Integer(maxIdle).intValue());

                        String maxWait = atts.getValue(tags.getTagById(conMaxWait));
                        logger.debug("     " + tags.getTagById(conMaxWait) + ": " + maxWait);
                        if(checkString(maxWait)) m_CurrentCPD.setMaxWait(new Integer(maxWait).intValue());

                        String minEvictableIdleTimeMillis = atts.getValue(tags.getTagById(conMinEvictableIdleTimeMillis));
                        logger.debug("     " + tags.getTagById(conMinEvictableIdleTimeMillis) + ": " + minEvictableIdleTimeMillis);
                        if(checkString(minEvictableIdleTimeMillis)) m_CurrentCPD.setMinEvictableIdleTimeMillis(new Long(minEvictableIdleTimeMillis).longValue());

                        String numTestsPerEvictionRun = atts.getValue(tags.getTagById(conNumTestsPerEvictionRun));
                        logger.debug("     " + tags.getTagById(conNumTestsPerEvictionRun) + ": " + numTestsPerEvictionRun);
                        if(checkString(numTestsPerEvictionRun)) m_CurrentCPD.setNumTestsPerEvictionRun(new Integer(numTestsPerEvictionRun).intValue());

                        String testOnBorrow = atts.getValue(tags.getTagById(conTestOnBorrow));
                        logger.debug("     " + tags.getTagById(conTestOnBorrow) + ": " + testOnBorrow);
                        if(checkString(testOnBorrow)) m_CurrentCPD.setTestOnBorrow(new Boolean(testOnBorrow).booleanValue());

                        String testOnReturn = atts.getValue(tags.getTagById(conTestOnReturn));
                        logger.debug("     " + tags.getTagById(conTestOnReturn) + ": " + testOnReturn);
                        if(checkString(testOnReturn)) m_CurrentCPD.setTestOnReturn(new Boolean(testOnReturn).booleanValue());

                        String testWhileIdle = atts.getValue(tags.getTagById(conTestWhileIdle));
                        logger.debug("     " + tags.getTagById(conTestWhileIdle) + ": " + testWhileIdle);
                        if(checkString(testWhileIdle)) m_CurrentCPD.setTestWhileIdle(new Boolean(testWhileIdle).booleanValue());

                        String timeBetweenEvictionRunsMillis = atts.getValue(tags.getTagById(conTimeBetweenEvictionRunsMillis));
                        logger.debug("     " + tags.getTagById(conTimeBetweenEvictionRunsMillis) + ": " + timeBetweenEvictionRunsMillis);
                        if(checkString(timeBetweenEvictionRunsMillis)) m_CurrentCPD.setTimeBetweenEvictionRunsMillis(new Long(timeBetweenEvictionRunsMillis).longValue());

                        String whenExhaustedAction = atts.getValue(tags.getTagById(conWhenExhaustedAction));
                        logger.debug("     " + tags.getTagById(conWhenExhaustedAction) + ": " + whenExhaustedAction);
                        if(checkString(whenExhaustedAction)) m_CurrentCPD.setWhenExhaustedAction(new Byte(whenExhaustedAction).byteValue());

                        String connectionFactoryStr = atts.getValue(tags.getTagById(connectionFactory));
                        logger.debug("     " + tags.getTagById(connectionFactory) + ": " + connectionFactoryStr);
                        if(checkString(connectionFactoryStr)) m_CurrentCPD.setConnectionFactory(connectionFactoryStr);
                        break;
                    }

                default :
                    {
                        // nop
                    }
            }
        }
        catch (Exception ex)
        {
            logger.error(ex);
            throw new PersistenceBrokerException(ex);
        }
    }

    private boolean checkString(String str)
    {
        return (str != null && !str.equals(""));
    }

    /**
     * endElement callback. most elements are build up from here.
     */
    public void endElement(String uri, String name, String qName)
    {
        try
        {
            switch (getLiteralId(qName))
            {
                case MappingRepository :
                    {
                        logger.debug(" < " + tags.getTagById(MappingRepository));
                        break;
                    }
                case RepositoryFragment :
                    {
                        logger.debug(" < " + tags.getTagById(RepositoryFragment));
                        break;
                    }
                case JdbcConnectionDescriptor :
                    {
                        logger.debug("   < " + tags.getTagById(JdbcConnectionDescriptor));
                        break;
                    }
                case connection_pool :
                    {
                        logger.debug("   < " + tags.getTagById(connection_pool));
                        break;
                    }
                case ClassDescriptor :
                    {
                        logger.debug("  < " + tags.getTagById(ClassDescriptor));
                        m_CurrentCLD = null;
                        break;
                    }
                case Class_extent :
                    {
                        break;
                    }

                case FIELDDESCRIPTOR :
                    {
                        logger.debug("    < " + tags.getTagById(FIELDDESCRIPTOR));
                        m_CurrentFLD = null;
                        break;
                    }
                case ReferenceDescriptor :
                    {
                        logger.debug("    < " + tags.getTagById(ReferenceDescriptor));
                        m_CurrentORD = null;
                        break;
                    }
                case FOREIGN_KEY :
                    {
                        logger.debug("    < " + tags.getTagById(FOREIGN_KEY));
                    }

                case CollectionDescriptor :
                    {
                        logger.debug("    < " + tags.getTagById(CollectionDescriptor));
                        m_CurrentCOD = null;
                        break;
                    }

                case Inverse_fk :
                    {
                        logger.debug("    < " + tags.getTagById(Inverse_fk));
                        break;
                    }

                case Fk_pointing_to_this_class :
                    {
                        logger.debug("    < " + tags.getTagById(Fk_pointing_to_this_class));
                        break;
                    }
                case Fk_pointing_to_items_class :
                    {
                        logger.debug("    < " + tags.getTagById(Fk_pointing_to_items_class));
                        break;
                    }

                    // handle failure:
                default :
                    {
                        logger.error("Ignoring unknown Element " + qName);
                    }
            }
        }
        catch (Exception ex)
        {
            logger.error(ex);
            throw new PersistenceBrokerException(ex);
        }
    }

    /**
     * characters callback.
     */
    public void characters(char ch[], int start, int length)
    {
        if (m_CurrentString == null)
            m_CurrentString = new String(ch, start, length);
        else
            m_CurrentString += new String(ch, start, length);
    }

    /**
     * Error callback.
     */
    public void error(SAXParseException e) throws SAXException
    {
        logger.error(e);
        throw e;
    }

    /**
     * fatal error callback.
     */
    public void fatalError(SAXParseException e) throws SAXException
    {
        logger.fatal(e);
        throw e;
    }

    /**
     * warning callback.
     */
    public void warning(SAXParseException e) throws SAXException
    {
        logger.warn(e);
        throw e;
    }

    /**
     * maps IsolationLevel literals to the corresponding id
     * @param isoLevel
     * @return the id
     */
    private int getIsoLevel(String isoLevel)
    {
        if (isoLevel.equalsIgnoreCase(LITERAL_IL_READ_UNCOMMITTED))
        {
            return IL_READ_UNCOMMITTED;
        }
        else if (isoLevel.equalsIgnoreCase(LITERAL_IL_READ_COMMITTED))
        {
            return IL_READ_COMMITTED;
        }
        else if (isoLevel.equalsIgnoreCase(LITERAL_IL_REPEATABLE_READ))
        {
            return IL_REPEATABLE_READ;
        }
        else if (isoLevel.equalsIgnoreCase(LITERAL_IL_SERIALIZABLE))
        {
            return IL_SERIALIZABLE;
        }
        else if (isoLevel.equalsIgnoreCase(LITERAL_IL_OPTIMISTIC))
        {
            return IL_OPTIMISTIC;
        }
        logger.warn("unknown isolation-level: " + isoLevel + " using RW_UNCOMMITTED as default");
        return defIsoLevel;
    }

    /**
     * returns a dynamic Proxy that implements all interfaces
     * implemented by the class described by cld.
     * @param cld the ClassDescriptor
     * @return Class the dynamically created proxy class
     */
    private Class getDynamicProxyClass(ClassDescriptor cld)
    {
        Class clazz = cld.getClassOfObject();
//         try
//         {
            return cld.getDynamicProxyClass();
//         }
//         catch (Throwable t)
//         {
//             // this error is not critical, just log a warning
//             logger.warn("can not use dynamic proxy for class " + clazz.getName() + ": " + t.getMessage());
//             return null;
//         }

    }
}
