package org.apache.ojb.broker.metadata.auto;

import org.apache.ojb.broker.metadata.CollectionDescriptor;
import org.apache.ojb.broker.metadata.DescriptorRepository;
import org.apache.ojb.broker.metadata.ClassDescriptor;
import org.apache.ojb.broker.metadata.FieldDescriptor;
import org.apache.ojb.broker.metadata.JdbcConnectionDescriptor;
import org.apache.ojb.broker.accesslayer.ConnectionFactoryDefaultImpl;
import org.apache.ojb.broker.accesslayer.ConnectionFactory;
import org.apache.ojb.broker.accesslayer.ConnectionManagerIF;
import org.apache.ojb.broker.accesslayer.LookupException;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerException;
import org.apache.ojb.broker.PersistenceBrokerSQLException;

import org.apache.ojb.broker.util.logging.LoggerFactory;
import org.apache.ojb.broker.util.logging.LoggingHelper;
import org.apache.ojb.broker.util.logging.Logger;

import java.util.Iterator;
import java.util.Map;
import java.util.Collection;
import java.util.HashSet;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import com.crossdb.sql.CreateTableQuery;
import com.crossdb.sql.SQLFactory;
import com.crossdb.sql.Column;

/**
 * This class will create the db automatically from the class descriptors
 *
 * @author Travis Reeder - travis@spaceprogram.com
 * @version 0.1
 * Date: Jul 8, 2002
 * Time: 10:31:16 AM
 *
 */
public class AutoCreatorDb {

	private PersistenceBroker broker;
	private Logger logger;
	private Collection handledClds = new HashSet();
			
	public AutoCreatorDb(PersistenceBroker broker) {
		this.broker = broker;
    	logger = LoggerFactory.getDefaultLogger();	
	}

	/**
	 * Obtain the appropriate CrossDB factory.
	 * TODO should we cache??
	 */
	private SQLFactory getCrossDBFactory(JdbcConnectionDescriptor jdbc) throws PersistenceBrokerException {
		String dbms = jdbc.getDbms();
		String crossdbDriver = getCrossdbDriverFromPlatform(dbms);
		SQLFactory factory= null;
		
		if (crossdbDriver == null) {
			throw new PersistenceBrokerException("cannot autocreate: no crossDB driver for : " + dbms);
		}
		logger.debug("Cross DB " + crossdbDriver + " for " + dbms);
		
		try {
			Class factory_class = Class.forName(crossdbDriver); // put the implementation string into forName()
			factory = (SQLFactory) (factory_class.newInstance());
		} catch (Exception ex) {
			logger.error("Cannot load crossDB", ex);
			throw new PersistenceBrokerException("error loading crossDB", ex);
		}
		return factory;
	}	
	
	private boolean isTablePresent(Connection c, String table) {
        PreparedStatement checkTable = null;
        boolean exists = false;
        
        //test if a table exists,  (copied from DBUtility)
        try {
   			checkTable = c.prepareStatement("Select * from " + table + " where 1=0");
			checkTable.executeQuery();
			exists = true;
        } catch (SQLException e) {
        }
        finally {
            if (checkTable != null) {
                try {
					checkTable.close();
				} catch (SQLException e) {
				}
            }
        }
		return exists;
	}

	/**
	 * Create tables for a single class if not already attmepted.
	 */
	public void createIfRequiredForClass(ClassDescriptor cd) throws PersistenceBrokerException {
		if (handledClds.contains(cd)) {
			return;
		}
		handledClds.add(cd);
		createForClass(cd);
	}
	
	/**
	 * Create tables for a single class.
	 */
	public void createForClass(ClassDescriptor cd) throws PersistenceBrokerException {
		
		// obtain connection
		JdbcConnectionDescriptor jdbcDescriptor	= cd.getConnectionDescriptor();
		
		ConnectionManagerIF connMgr = broker.getConnectionManager();
		
		Connection conn = null;
		try {
			// shouldn't we use connMgr.getConnectionForClass()? but is deprecated!		
			conn = connMgr.getConnection(); 
		} catch (LookupException ex) {
			throw new PersistenceBrokerException("Unable to obtain connection");
		}
		SQLFactory factory = getCrossDBFactory(jdbcDescriptor);
		
		// check if table exists
		String tableName = cd.getFullTableName();
		if (tableName == null) return;
		
		if (isTablePresent(conn, tableName)) {
			logger.debug("Table " + tableName + " exists");
			return;
		}
		
		// create table
		logger.debug("Attempting auto table creation for " + tableName);
		CreateTableQuery cq = factory.getCreateTableQuery();
		cq.setName(tableName);
		
		// now get columns
		FieldDescriptor fd[] = cd.getFieldDescriptions();
		if (fd != null) {
			for (int i = 0; i < fd.length; i++) {
				FieldDescriptor descriptor = fd[i];
				Column c = new Column(descriptor.getColumnName(), descriptor.getColumnJdbcType());
				c.setAutoIncrement(descriptor.isAutoIncrement());
				if (descriptor.getLength() != 0) {
					c.setSize(descriptor.getLength());
				}
				cq.addColumn(c);
			}
		}
		try {
			cq.execute(conn);
			logger.info("Table " + tableName + " created");
		}
		catch (SQLException ex) {
			logger.error("cannont create Table " + tableName, ex);
			throw new PersistenceBrokerSQLException("cannot create " + tableName, ex);
		}
		
		// create any required indirection tables
		Collection colds = cd.getCollectionDescriptors();
		for (Iterator it = colds.iterator(); it.hasNext();) {
			CollectionDescriptor cold = (CollectionDescriptor)it.next();
			if (cold.isMtoNRelation()) {
				String indirectTable = cold.getIndirectionTable();
				
				//QUESTION: how do we handle multiple FKs?
				// In the code below we assume that getFksToThisClass() returns the
				// same number of PKs as getPkFields() and in the same order.
				// Is this correct?
				// The unit tests currently only check the single FK case.				
				Object[] fkColForThis = cold.getFksToThisClass();
				Object[] fkColForItem = cold.getFksToItemClass();
				
				FieldDescriptor[] pksForThis = cd.getPkFields();
				ClassDescriptor cdItem = broker.getClassDescriptor(cold.getItemClass());
				FieldDescriptor[] pksForItem = cdItem.getPkFields();
				
				cq = factory.getCreateTableQuery();
				cq.setName(indirectTable);
				addColumnsToTableCreate(fkColForThis, pksForThis, cq);
				addColumnsToTableCreate(fkColForItem, pksForItem, cq);

				try {
					cq.execute(conn);
					logger.info("IndirectionTable " + indirectTable + " created");
				}
				catch (SQLException ex) {
					logger.error("cannont create Table " + indirectTable, ex);
					throw new PersistenceBrokerSQLException("cannot create " + indirectTable, ex);
				}				
			}
		}
	}

	private void addColumnsToTableCreate(Object[] cols, FieldDescriptor[] pks, CreateTableQuery cq) {
		int nbCols = cols.length;
		int nbPks = pks.length;

		if (nbCols != nbPks) {
			logger.warn("Column, PK length mismatch! for " + cq);
		}
	
		for (int i=0; i<Math.min(nbCols, nbPks); i++) {
			Column c = new Column(cols[i].toString(), pks[i].getColumnJdbcType());
			cq.addColumn(c);
		}
	}
	
	/**
	 *  Autocreate all tables in a repository.
	 */
	public void createForRepository(DescriptorRepository repos) {
			Map descriptorTable = repos.getDescriptorTable();
			for (Iterator it = descriptorTable.values().iterator(); it.hasNext();) {

				ClassDescriptor cld = (ClassDescriptor) it.next();
				// now create tables based on this
				createForClass(cld);
			}
	}
		
	/**
	 * TODO: Don't know why the Pbroker is needed for the newConnection() method, but since it is
	 * i've passed it in here.
	 */
	public static void autoCreate(DescriptorRepository repos, PersistenceBroker pb) {
		AutoCreatorDb creator = new AutoCreatorDb(pb);
		creator.createForRepository(repos);
	}

	/**
	 * Obtain the cross driver for the given database platform.
	 * Currently supported platforms : <UL>
	 * <LI>hsqldb => com.spaceprogram.sql.hsqldb.HsqldbFactory
	 * <LI>mysql => com.spaceprogram.sql.mysql.MySQLFactory
	 * <LI>oracle => com.thinkvirtual.sql.oracle.OracleSQLFactory
	 * <LI>mssql => com.thinkvirtual.sql.sqlserver.SQLServerFactory
	 * <LI>sybase =>com.thinkvirtual.sql.sybase.SybaseFactory
	 * </UL>
	 */
	static String getCrossdbDriverFromPlatform(String platform) {
		String ret = null;
		if (platform.equalsIgnoreCase("hsqldb")) {
			ret = "com.spaceprogram.sql.hsqldb.HsqldbFactory";

		}
		else if (platform.equalsIgnoreCase("mysql")) {
			ret = "com.spaceprogram.sql.mysql.MySQLFactory";

		}
		else if (platform.equalsIgnoreCase("mssql")) {
			ret = "com.thinkvirtual.sql.sqlserver.SQLServerFactory";
		}
		else if (platform.equalsIgnoreCase("oracle")) {
			ret = "com.thinkvirtual.sql.oracle.OracleSQLFactory";
		}
		else if (platform.equalsIgnoreCase("sybase")) {
			ret = "com.thinkvirtual.sql.sybase.SybaseFactory";
		}
		return ret;


	}

}
